@@ -10,9 +10,9 @@
 
 import os
 import sys
-import zipimport
 
-FILE_NAME = sys.executable
+import BUILD_CONSTANTS
+
 DIR_NAME = os.path.dirname(sys.executable)
 
 paths = os.environ.get("LD_LIBRARY_PATH", "").split(os.pathsep)
@@ -25,19 +25,21 @@ if DIR_NAME not in paths:
 sys.frozen = True
 sys.path = sys.path[:4]
 
-os.environ["TCL_LIBRARY"] = os.path.join(DIR_NAME, "tcl")
-os.environ["TK_LIBRARY"] = os.path.join(DIR_NAME, "tk")
+if hasattr(BUILD_CONSTANTS, "TCL_LIBRARY"):
+    os.environ["TCL_LIBRARY"] = os.path.join(DIR_NAME,
+                                             BUILD_CONSTANTS.TCL_LIBRARY)
 
-m = __import__("__main__")
-importer = zipimport.zipimporter(os.path.dirname(os.__file__))
-name, ext = os.path.splitext(os.path.basename(os.path.normcase(FILE_NAME)))
-moduleName = "%s__main__" % name
-code = importer.get_code(moduleName)
-exec(code, m.__dict__)
+if hasattr(BUILD_CONSTANTS, "TK_LIBRARY"):
+    os.environ["TK_LIBRARY"] = os.path.join(DIR_NAME,
+                                            BUILD_CONSTANTS.TK_LIBRARY)
 
-versionInfo = sys.version_info[:3]
-if versionInfo >= (2, 5, 0) and versionInfo <= (2, 6, 4):
-    module = sys.modules.get("threading")
-    if module is not None:
-        module._shutdown()
+if hasattr(BUILD_CONSTANTS, "PYTZ_TZDATADIR"):
+    os.environ["PYTZ_TZDATADIR"] = os.path.join(DIR_NAME,
+                                                BUILD_CONSTANTS.PYTZ_TZDATADIR)
 
+def run():
+    name = __name__.rpartition("__init__")[0] + "__main__"
+    code = __loader__.get_code(name)
+    m = __import__('__main__')
+    m.__dict__['__file__'] = code.co_filename
+    exec(code,  m.__dict__)
